#ifndef _LOCK_SLEEP_H
#define _LOCK_SLEEP_H

#ifdef __SMP__

/* Locks */
#if defined(linux)

#include <sys/types.h>
#include <asm/bitops.h>

#define	LOCK_BIT	1

#if defined(__KERNEL__)
typedef int lock_t;
#else
typedef	struct _locky {
	pid_t		holder;
	u_int32_t	lock;
} lock_t;
#endif

#define	TestAndSetLock(a)	set_bit(LOCK_BIT, &(a)->lock)
#define TestLock(a)		test_bit(LOCK_BIT, &(a)->lock)
#define	ClearLock(a)		clear_bit(LOCK_BIT, &(a)->lock)

#if defined(__KERNEL__)
#define	BlockAll()	cli()
#define	UnBlockAll()	sti()
#else
extern lock_t *dummy_lock;
#define	BlockAll()	{while (TestAndSetLock(dummy_lock)) {/*printf("%d: blocked by %d!\n", getpid(), dummy_lock->holder);*/ sleep(1);} dummy_lock->holder=getpid();}
#define UnBlockAll()	(void)ClearLock(dummy_lock)
#endif /* !__KERNEL__ */

#endif /* linux */

/* Beds */
#if defined(linux)

#if defined(__KERNEL__)
typedef struct wait_queue *bed_t;
#define SleepOn(bed, limit)	interruptible_sleep_on(&bed)
#define RouseFrom(bed)	wake_up_interruptible(&bed)
#else
typedef lock_t bed_t;
#ifdef linux
extern __inline__ int
SleepOn(bed_t *bed, int limit)
{ 
	int sleep_count;

	(void)TestAndSetLock(bed);
	UnBlockAll();
	for (sleep_count = 0;
		 TestLock(bed) && sleep_count < limit; ++sleep_count)
			 sleep(1);
	 BlockAll();
	 return sleep_count >= limit ? -1 : 0;
}
#else
#define SleepOn(bed, limit)    \
        { (void)TestAndSetLock(bed); UnBlockAll(); while (TestLock(bed)) sleep(1); BlockAll(); }
#endif

#define RouseFrom(bed)	ClearLock(bed)
#endif

#endif

#ifdef linux
extern __inline__ void
SleepyLock(lock_t *the_lock, bed_t *the_bed, char *message)
{
	BlockAll();
	while (TestAndSetLock(the_lock)) {
		SleepOn(the_bed, 20);
	}
	UnBlockAll();
}

extern __inline__ void
SleepyTest(lock_t *the_lock, bed_t *the_bed, char *message)
{
	BlockAll();
	while (TestLock(the_lock)) {
		SleepOn(the_bed, 20);
	}
	UnBlockAll();
}

extern __inline__ void
SleepyUnLock(lock_t *the_lock, bed_t *the_bed)
{
	BlockAll();
	ClearLock(the_lock);
	RouseFrom(the_bed);
	UnBlockAll();
}
#endif

#endif /* __SMP__ */

#endif /* _LOCK_SLEEP_H */
